
#include "qobjects.h"

#include <ctime>

int main(int argc, char* argv[]){
   
//------------------------------initialize parameters

   const int seed = 12356;                  // random seed

   const double hbar = 1,                   // fundamental constants
                e  = 2.7182818,
                pi = 3.1415927;
		
   const std::complex<double> ei(0,1); 
   
   const unsigned int Ti =  500;            // time parameters
   
   const unsigned int nx = 128,             // size parameters
		              ns = 10000*Ti,       
					  rep_out = 10000,
					  ser_out = 10,
					  state_out = 4000;  
   
   const double L = std::sqrt(nx*2*pi),                    
                dx = L/nx,
                dp = 2*pi*hbar/L,
                Lp = nx*dp,
   
                T(Ti),
                dt = T/ns,
				tstart = 300, 
				tfin = T,
   
                meanx0 = 0,             // initial state parameters
                meanp0 = 0,
                Vx0 = hbar/2,
                Vp0 = hbar*hbar/(4*Vx0), 

                k = 0.1,    
                w = 2*pi*10,   
				A4 = 0.00,  
				A6 = 0.00;   
		
                std::cout << "(LX,Lp) = (" << Lp << "," << L << ")"<< std::endl;
		
//--------------------------------initialize grids and functions of x and p

   std::vector<double> x(nx),p(nx),Fx2(nx),Fp2(nx),Fx4(nx),Fx6(nx);
   std::vector< std::complex<double> > FUx(nx),FUp(nx);
                              
   // x and functions of x (with offset)										            
   for (unsigned int i=0;i<nx;++i) { x[i] = -L/2 + i*dx; }
   
   // p and functions of p ( with offset in fftshifted space)
   for (unsigned int i=nx/2;i<nx;++i) { p[i-nx/2] = -Lp/2 + i*dp; }
   for (unsigned int i=0;i<nx/2;++i)  { p[i+nx/2] = -Lp/2 + i*dp; }

//--------------------------------initialise files

   std::ofstream ostates_r("states_real.dat");
   std::ofstream ostates_i("states_imag.dat");
   std::ofstream oseries("series.dat"); 
 
//--------------------------------initialise the state psi
   
   qstate psi("Gaussian",nx,L,hbar,Vx0,Vp0,meanx0,meanp0),
          d_psi = psi,
          i_psi = psi;
		  
//--------------------------------initialize the operators

   qoperator Ix("identity",nx);
   
// ------------------------------- fftw initialization

   fftw_plan plan1, plan2;
   psi.fftw_init(plan1, plan2);

//--------------------------------generate the noise (dW's)

   std::vector<double> noise;
   double r8kdt = std::sqrt(8*k*dt);
   double tn1,tn2,rn;
   srand(seed);
   for (unsigned int i=0;i<(ns*3+1)/2;++i) {
     rn = 2;
     while (rn >= 1 || rn == 0) {
            tn1 = 2*static_cast<double>(rand())/static_cast<double>(RAND_MAX) - 1;
            tn2 = 2*static_cast<double>(rand())/static_cast<double>(RAND_MAX) - 1;
            rn = tn1*tn1 + tn2*tn2;
     }
     rn = std::sqrt((-2*std::log(rn)/rn));
     noise.push_back(tn1*rn/r8kdt);
     noise.push_back(tn2*rn/r8kdt);
   }
   
//------------------------------------------------------------
//                        Simulation
//------------------------------------------------------------
   
   double t=0,               // initialise variables
          dr =0,
          dW =0,
          meanx = meanx0,
		  meanp = meanp0,
		  Mx2 = Vx0 + meanx0*meanx0,
	      Vp = Vp0,
          Vx = Vx0;
	 
   time_t time0, *time_ptr = 0;       // timing initialisation
   time0 = time(time_ptr);
   
   oseries << t << " " << noise[0]*r8kdt << " " << meanx  << " " << meanp  << " " 
                       << Vx       << " " << Vp     << std::endl;
					   
   psi.out_to_2files(ostates_r, ostates_i);


   for (int i=0;i<ns;++i) {

     t = i*dt;
     if (std::div(i,rep_out).rem==0) {std::cout << i << std::endl;}

     //--------------------------------- construct the operators

     for (unsigned int j=0;j<nx;++j) {     // Potential and Measurement
       Fx2[j] = x[j]*x[j];
	   Fx4[j] = Fx2[j]*Fx2[j];
	   Fx6[j] = Fx4[j]*Fx2[j];
       FUx[j] = std::exp(  (-ei*dt/hbar)*( 0.5*w*Fx2[j] + A4*Fx4[j] + A6*Fx6[j] ) 
						  +(-k*dt)*( Fx4[j] - 2*Fx2[j]*(Mx2 + noise[i]) )  );
     }
                                         
     for (unsigned int j=0;j<nx;++j) {     // functions of p (in fftshifted space)
       Fp2[j] = p[j]*p[j];
       FUp[j] = std::exp(-ei*0.5*w*Fp2[j]*dt/hbar);
     }

     qoperator Xx(x),Xx2(Fx2),Ux(FUx);
     qoperator Pp(p),Pp2(Fp2),Up(FUp);

     //---------------------------------- evolve 

     psi.apply(Ux);
     psi.fftw(plan1); 
     psi.apply(Up);
     psi.normalise();

     meanp = (Pp.expect(psi)).real();
     Vp = (Pp2.expect(psi)).real() - meanp*meanp;
     
     psi.fftw(plan2);
     psi.normalise(); 

     meanx = (Xx.expect(psi)).real(); 
	 Mx2 = (Xx2.expect(psi)).real(); 
     Vx = Mx2 - meanx*meanx; 

     //---------------------------------- output
	 
     if (std::div(i+1,state_out).rem==0 && t>=tstart && t<=tfin) { psi.out_to_2files(ostates_r, ostates_i); }
     
	 if (std::div(i+1,ser_out).rem==0) {
       oseries << t+dt << " " << noise[i]*r8kdt << " " << meanx  << " " << meanp  << " " 
                              << Vx     << " " << Vp     << std::endl;
	 }

   }

   std::cout << "Simulation took " << difftime(time(time_ptr),time0) << " seconds" << std::endl;
   psi.fftw_fin(plan1,plan2);

   // tidy up 
   psi.fftw_fin(plan1,plan2);
   ostates_r.close();
   ostates_i.close();
   oseries.close();
   
   return 1;
}
